# https://github.com/python-poetry/poetry/pull/2366#issuecomment-652418094
from importlib import metadata

from .models import *  # noqa: F403
from .utils import json_dumps

__version__ = metadata.version(__name__)

__all__ = [  # noqa: F405
    # Core Models
    "CalcType",
    "Model",
    "Provenance",
    "Structure",
    "Molecule",
    "Files",
    "FileInput",
    "ProgramInput",
    "ProgramOutput",
    "DualProgramInput",
    "SinglePointResults",
    "OptimizationResults",
    "SinglePointOutput",
    "Wavefunction",
    "ProgramArgs",
    "ProgramArgsSub",
    "json_dumps",
]



"""Input models for quantum chemistry calculations."""

import warnings
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, field_serializer
from typing_extensions import Self, TypeVar

from .base_models import CalcType, Files, Model
from .structure import Structure

__all__ = [
    "FileInput",
    "ProgramInput",
    "DualProgramInput",
    "ProgramArgs",
    "Inputs",
    "InputType",
    "StructuredInputs",
    "ProgramArgsSub",
]


class FileInput(Files):
    """File and command line argument inputs for a calculation.

    Attributes:
        files Files: A dict mapping filename to str or bytes data.
        cmdline_args: A list of command line arguments to be passed to the program.
        extras Dict[str, Any]: Additional information to bundle with the object. Use for
            schema development and scratch space.
    """

    cmdline_args: List[str] = []

    @classmethod
    def from_directory(cls, directory: Union[Path, str], **kwargs) -> Self:
        """Create a new FileInput and collect all files in the directory."""
        obj = cls(**kwargs)
        directory = Path(directory)
        obj.add_files(directory)
        return obj


class _KeywordsMixin(BaseModel):
    keywords: Dict[str, Any] = {}


class _StructureKeywordsMixin(_KeywordsMixin):
    """
    Attributes:
        structure: The structure to be used in the calculation or SMILES string.
    """

    structure: Structure

    def __init__(self, **data: Any):
        """Backwards compatibility for 'molecule' attribute."""

        # TODO: Remove in future versions.
        if "molecule" in data:
            warnings.warn(
                "Use of 'molecule' attribute is deprecated. Use 'structure' instead.",
                FutureWarning,
                stacklevel=2,
            )
            data["structure"] = data.pop("molecule")
        super().__init__(**data)

    @property
    def molecule(self) -> Structure:
        """Backwards compatibility for 'molecule' attribute."""
        warnings.warn(
            "Use of 'molecule' attribute is deprecated. Use 'structure' instead.",
            FutureWarning,
            stacklevel=2,
        )
        return self.structure


class ProgramArgs(FileInput, _KeywordsMixin):
    """Generic arguments for a program without a calctype or structure specification.

    This class is used by `DualProgramInput` or multi-step calculations to
    specify `subprogram_args` or a basic program arguments multistep algorithm in
    BigChem. It is not intended to be used directly for single-step calculations since
    it lacks a `calctype` and `structure`.


    Attributes:
        model Model: The model for the quantum chemistry calculation.
        keywords Dict[str, Any]: A dict of keywords to be passed to the program
            excluding model and calctype. Defaults to an empty dict.
        files Files: Files to be passed to the QC program.
        extras Dict[str, Any]: Additional information to bundle with the object. Use
            for schema development and scratch space.
    """

    model: Model


class ProgramArgsSub(FileInput, _KeywordsMixin):
    """Generic arguments for a program that also calls a subprogram.

    This class is needed for multi-step calculations where the calctype and structure
    are specified only once for the entire calculation, e.g., multistep_opt in BigChem.

    Attributes:
        keywords: A dict of keywords to be passed to the program excluding model and
            calctype. Defaults to an empty dict.
        files: Files to be passed to the QC program.
        model: The model for the quantum chemistry calculation
        subprogram: The name of the subprogram to use.
        subprogram_args: The ProgramArgs for the subprogram.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    model: Optional[Model] = None
    subprogram: str
    subprogram_args: ProgramArgs


class ProgramInput(ProgramArgs, _StructureKeywordsMixin):
    """Input for a single quantum chemistry program. This is the most common input type.

    Attributes:
        calctype CalcType: The type of calculation to perform.
        model Model: The model for the quantum chemistry calculation.
        keywords Dict[str, Any]: A dict of keywords to be passed to the program
            excluding model and calctype. Defaults to an empty dict.
        structure Structure: The structure to be used in the calculation.
        files Files: Files to be passed to the QC program.
        extras Dict[str, Any]: Additional information to bundle with the object. Use
            for schema development and scratch space.

    Example:
        ```python
        from qcio.models import ProgramInput, Structure

        struct = Structure.open("path/to/structure.xyz")

        prog_inp = ProgramInput(
            calctype = "energy",
            structure = struct,
            model = {"method": "hf", "basis": "6-31G"},
            keywords = {"maxsteps": "250"},  # Optional
            files = {"file1": b"binary data"}  # Optional
        )
        ```
    """

    calctype: CalcType

    @field_serializer("calctype")
    def _serialize_calctype(self, calctype: CalcType, _info) -> str:
        """Serialize CalcType to string"""
        return calctype.value


class DualProgramInput(ProgramArgsSub, ProgramInput):
    """Input for a two program calculation.

    Attributes:
        calctype CalcType: The type of calculation to perform.
        model Model: The model for the quantum chemistry calculation.
        keywords Dict[str, Any]: A dict of keywords to be passed to the program
            excluding model and calctype. Defaults to an empty dict.
        structure Structure: The structure to be used in the calculation.
        files Files: Files to be passed to the QC program.
        subprogram: The name of the subprogram to use.
        subprogram_args ProgramArgs: The ProgramArgs for the subprogram.
        extras Dict[str, Any]: Additional information to bundle with the object. Use
            for schema development and scratch space.

    Example:
        ```python
        from qcio.models import DualProgramInput, Structure

        struct = Structure.open("path/to/structure.xyz")

        prog_inp = DualProgramInput(
            calctype = "optimization",
            structure = struct,
            keywords = {"maxiter": "250"},  # Optional
            subprogram = "orca",
            subprogram_args = ProgramArgs(
                model = {"method": "wb97x-d3", "basis": "def2-SVP"},
                keywords = {"convthre": "1e-6"},  # Optional
            )
        )
        ```
    """


Inputs = Union[FileInput, ProgramInput, DualProgramInput]
InputType = TypeVar("InputType", bound=Inputs)
StructuredInputs = Union[ProgramInput, DualProgramInput]



import warnings
from collections import Counter
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Tuple, Union

import numpy as np
from pydantic import field_serializer, model_validator
from typing_extensions import Self

from qcio.constants import BOHR_TO_ANGSTROM
from qcio.constants import periodic_table as pt
from qcio.helper_types import SerializableNDArray

from .base_models import QCIOModelBase
from .utils import renamed_class, smiles_to_structure, structure_to_smiles

if TYPE_CHECKING:
    from pydantic.typing import ReprArgs

__all__ = ["Structure", "Identifiers", "Molecule", "DistanceUnits"]


class DistanceUnits(str, Enum):
    """Distance units for the Structure.distance method.

    Attributes:
        bohr (str): The distance in Bohr.
        angstrom (str): The distance in Angstrom.

    """

    bohr = "bohr"
    angstrom = "angstrom"


class Identifiers(QCIOModelBase):
    """Structure identifiers.

    Attributes:
        name: A human-readable, common name for the structure.
        name_IUPAC: The IUPAC name of the structure.
        smiles: The SMILES representation of the structure.
        canonical_smiles: The canonical SMILES representation of the structure.
        canonical_smiles_program: The program used to generate the canonical SMILES.
        canonical_explicit_hydrogen_smiles: The canonical explicit hydrogen SMILES
            representation of the structure.
        canonical_isomeric_smiles: The canonical isomeric SMILES representation of the
            structure.
        canonical_isomeric_explicit_hydrogen_smiles: The canonical isomeric explicit
            hydrogen SMILES representation of the structure.
        canonical_isomeric_explicit_hydrogen_mapped_smiles: The canonical isomeric
            explicit hydrogen mapped SMILES representation of the structure.
        inchi: The InChI representation of the structure.
        inchikey: The InChIKey representation of the structure.
        pubchem_cid: The PubChem Compound ID of the structure.
        pubchem_sid: The PubChem Substance ID of the structure.
        pubchem_conformerid: The PubChem Conformer ID of the structure.
        extras Dict[str, Any]: Additional information to bundle with the object. Use for
            schema development and scratch space.
    """

    name: Optional[str] = None
    name_IUPAC: Optional[str] = None
    smiles: Optional[str] = None
    canonical_smiles: Optional[str] = None
    canonical_smiles_program: Optional[str] = None
    canonical_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_smiles: Optional[str] = None
    canonical_isomeric_explicit_hydrogen_mapped_smiles: Optional[str] = None
    inchi: Optional[str] = None
    inchikey: Optional[str] = None
    pubchem_cid: Optional[str] = None
    pubchem_sid: Optional[str] = None
    pubchem_conformerid: Optional[str] = None


class IDs2(Identifiers):
    """Second one"""


# class Bond(QCIOModelBase):
#     """A bond object.

#     Attributes:
#         indices: The indices of the atoms in the bond.
#         order: The order of the bond.
#     """

#     indices: Tuple[int, int]
#     order: Optional[float]


class Structure(QCIOModelBase):
    """A Structure object with atoms and their corresponding cartesian coordinates,
        charge, multiplicity, and identifiers such as name, smiles, etc.

    Attributes:
        symbols: The atomic symbols of the structure.
        geometry: The geometry of the structure in Cartesian coordinates. Units are
            Bohr (AU).
        identifiers: Identifiers for the structure such as name, smiles, etc.
        charge: The molecular charge.
        multiplicity: The molecular multiplicity.
        connectivity: Explicit description of the bonds between atoms. Each tuple
            contains the indices of the atoms in the bond and the order of the bond.
            E.g., `[(0, 1, 1.0), (1, 2, 2.0)]` indicates a single bond between atoms 0
            and 1 and a double bond between atoms 1 and 2.
        extras Dict[str, Any]: Additional information to bundle with the object. Use
            for schema development and scratch space.
        ids: `@property` Shortcut to access identifiers.
        geometry_angstrom: `@property` The geometry of the structure in Angstrom.
        atomic_numbers: `@property` The atomic numbers of the atoms in the structure.
        formula: `@property` The molecular formula of the structure using the Hill
            System.
    """

    symbols: List[str]
    geometry: SerializableNDArray  # Coerced to 2D array
    charge: int = 0
    multiplicity: int = 1
    identifiers: Identifiers = Identifiers()
    connectivity: List[Tuple[int, int, float]] = []
    _xyz_comment_key: ClassVar[str] = "xyz_comments"

    def __init__(self, **data: Any):
        """Create a new Structure object.

        Example:
            ```python
            from qcio import Structure

            structure = Structure(
                symbols=["H", "O", "H"],
                geometry=[[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0]],
                charge=0,  # optional; defaults to 0
                multiplicity=1,  # optional; defaults to 1
                identifiers={"smiles": "CCO"},  # optional
            )

            ```
        """
        # Backwards compatibility for 'ids' attribute.
        if identifiers := data.pop("ids", None):
            warnings.warn(
                "Passing 'ids' is deprecated and will be removed in a future "
                "release. Please use 'identifiers' instead. Once instantiated, "
                "you can use structure.ids to access the identifiers as a shortcut.",
                category=FutureWarning,
                stacklevel=2,
            )
            data["identifiers"] = identifiers
        super().__init__(**data)

    @classmethod
    def open(
        cls,
        filepath: Union[Path, str],
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
    ) -> Union["Structure", List["Structure"]]:
        """Open a structure or structures from a file.

        Args:
            filepath: The path to the file to open. May be a path to a formerly saved
                Structure file or an XYZ file.
            charge: The molecular charge of the structure. Only used when opening an XYZ
                file and if not set in the file.
            multiplicity: The molecular multiplicity of the structure. Only used when
                opening an XYZ file and if not set in the file.

        Returns:
            A Structure object or a list of Structure objects for a multi-structure XYZ
                file.

        Example:
            ```python
            struct = Structure.open("path/to/structure.json")
            ```

            ```python
            struct = Structure.open("path/to/structure.xyz", charge=-1, multiplicity=3)
            ```

            ```python
            structures = Structure.open("path/to/structures.xyz")
            ```
        """
        filepath = Path(filepath)

        if (charge or multiplicity) and filepath.suffix != ".xyz":
            raise ValueError(
                "Charge and multiplicity can only be set when opening an XYZ file."
            )

        if filepath.suffix == ".xyz":
            xyz_str = filepath.read_text()
            structures = cls.from_xyz_multi(
                xyz_str, charge=charge, multiplicity=multiplicity
            )
            if len(structures) == 1:
                return structures[0]
            return structures

        return super().open(filepath)

    def save(
        self,
        filepath: Union[Path, str],
        exclude_none: bool = True,
        exclude_unset: bool = True,
        indent: int = 4,
        **kwargs,
    ) -> None:
        """Save a Structure to a file.

        Args:
            filepath: The path to save the structure to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            exclude_unset: If True, attributes that have not been set will not be
                written to the file.
            indent: The number of spaces to use for indentation in the JSON file. 0
                creates a more compact JSON file, 4 is more human-readable.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Notes:
            If the filepath has a `.xyz` extension, the structure will be saved to an
            XYZ file. qcio will automatically convert the geometry to Angstrom and add
            the charge and multiplicity to the comments line of the XYZ file.

        Example:
            ```python
            struct.save("path/to/structure.json")
            ```

            ```python
            struct.save("path/to/structure.xyz")
            ```
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            filepath.write_text(self.to_xyz(**kwargs))
            return
        super().save(filepath, exclude_none, exclude_unset, indent, **kwargs)

    @property
    def ids(self) -> Identifiers:
        """Shortcut to access the identifiers."""
        return self.identifiers

    @classmethod
    def from_smiles(
        cls,
        smiles: str,
        *,
        program: str = "rdkit",
        force_field: str = "MMFF94s",
        multiplicity: int = 1,
    ) -> Self:
        """Create a new Structure object from a SMILES string.

        Args:
            smiles: The SMILES string.
            program: The program to use for the conversion. Defaults to "rdkit".
            force_field: The force field to use. E.g., UFF, MMFF94, MMFF94s, etc.
            multiplicity: The multiplicity of the structure.

        Returns:
            A Structure object with identifiers for SMILES and canonical SMILES.

        Example:
            ```python
            struct = Structure.from_smiles("CN1C=NC2=C1C(=O)N(C(=O)N2C)C")

            print(struct.ids.smiles)
            # Output: 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'

            print(struct.ids.canonical_smiles)
            # Output: 'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'
            ```
        """
        dict_repr = smiles_to_structure(smiles, program, force_field)
        dict_repr["multiplicity"] = multiplicity
        return cls(**dict_repr)

    @classmethod
    def from_xyz(
        cls,
        xyz_str: str,
        *,
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
    ) -> Self:
        """Create a Structure from an XYZ file or string.

        Args:
            xyz_str: The XYZ string.
            charge: The molecular charge of the structure. If not provided, will read
                from the XYZ string if set or default to 0.
            multiplicity: The molecular multiplicity of the structure. If not provided,
                will read from the XYZ string if set or default to 1.

        Note:
            Will read qcio data such as `charge` and `multiplicity` from the comments
            line with a `qcio_key=value` format (if it is present). Also will read in
            qcio__identifiers_* keys and additional non-qcio comments.

        Example:
            ```python
            struct = Structure.from_xyz(xyz_str)
            ```
        """

        lines = xyz_str.split("\n")

        num_atoms = int(lines[0])

        # Collect comments
        structure_kwargs: Dict[str, Any] = {}
        identifier_kwargs: Dict[str, Any] = {}
        other_comments: List[str] = []

        for item in lines[1].strip().split():
            if item.startswith("qcio__identifiers_"):
                key = item.split("=")[0].replace("qcio__identifiers_", "")
                value = item.split("=")[1]
                identifier_kwargs[key] = value
            elif item.startswith("qcio_"):
                key = item.split("=")[0].replace("qcio_", "")
                value = item.split("=")[1]
                structure_kwargs[key] = value
            else:
                other_comments.append(item)

        if charge is not None and "charge" in structure_kwargs:
            raise ValueError("Charge cannot be set in the file and as an argument.")
        if multiplicity is not None and "multiplicity" in structure_kwargs:
            raise ValueError(
                "Multiplicity cannot be set in the file and as an argument."
            )

        # Set charge and multiplicity if provided
        if charge is not None:
            structure_kwargs["charge"] = charge
        if multiplicity is not None:
            structure_kwargs["multiplicity"] = multiplicity

        symbols = []
        geometry = []
        for line in lines[2 : 2 + num_atoms]:
            split_line = line.split()
            symbols.append(split_line[0])
            geometry.append([float(val) / BOHR_TO_ANGSTROM for val in split_line[1:]])

        return cls(
            symbols=symbols,
            geometry=geometry,
            **structure_kwargs,
            identifiers=Identifiers(**identifier_kwargs),
            extras={cls._xyz_comment_key: other_comments},
        )

    @classmethod
    def from_xyz_multi(
        cls,
        xyz_str: str,
        charge: Optional[int] = None,
        multiplicity: Optional[int] = None,
    ) -> List["Structure"]:
        """Parse a multi-structure XYZ file into a list of Structure objects.

        Args:
            xyz_str: The multi-structure XYZ string.

        Returns:
            A list of Structure objects.
        """
        lines = xyz_str.strip().split("\n")
        structures = []
        i = 0

        while i < len(lines):
            num_atoms = int(lines[i])
            structure_block = "\n".join(lines[i : i + num_atoms + 2])
            structure = cls.from_xyz(
                structure_block, charge=charge, multiplicity=multiplicity
            )
            structures.append(structure)
            i += num_atoms + 2
            while i < len(lines) and not lines[i].strip():
                i += 1  # Skip any empty lines between structures

        return structures

    def distance(
        self, i: int, j: int, units: DistanceUnits = DistanceUnits.bohr
    ) -> float:
        """Calculate the distance between two atoms.

        Args:
            i: The index of the first atom.
            j: The index of the second atom.
            units: The units to return the distance in. Defaults to "bohr".
                May be "bohr" or "angstrom".

        Returns:
            The distance between the atoms in units (Bohr or Angstrom).

        Example:
            ```python
            struct.distance(0, 1)
            1.34
            ```
        """
        distance = np.linalg.norm(self.geometry[i] - self.geometry[j])
        if units == DistanceUnits.angstrom:
            return float(distance * BOHR_TO_ANGSTROM)
        return float(distance)

    def to_smiles(self, program: str = "rdkit", hydrogens: bool = False) -> str:
        """Generate the canonical SMILES representation of the structure.

        Args:
            program: The program to use for the conversion. Defaults to "rdkit".
            hydrogens: Whether to include hydrogens in the SMILES string. Defaults to
                False.

        Returns:
            The canonical SMILES representation of the structure.

        Example:
            ```python
            struct.to_smiles()
            'CN1C=NC2=C1C(=O)N(C(=O)N2C)C'
            ```
        """
        return structure_to_smiles(self, program=program, hydrogens=hydrogens)

    def to_xyz(self, precision: int = 17) -> str:
        """Return an xyz string representation of the structure.

        Args:
            precision: The number of decimal places to include in the xyz file. Default
                17 which captures all precision of float64.
        Notes:
            Will add qcio data such as charge and multiplicity to the comments line with
            a `qcio_key=value` format.
        """

        qcio_data = {  # These get added to comments line (line 2) in xyz file
            "qcio_charge": self.charge,
            "qcio_multiplicity": self.multiplicity,
        }

        # Add identifiers to qcio_data
        for key, value in self.identifiers.__dict__.items():
            if key != "extras" and value:
                qcio_data[f"qcio__identifiers_{key}"] = value

        assert isinstance(self.geometry, np.ndarray)  # For mypy
        geometry_angstrom = self.geometry * BOHR_TO_ANGSTROM

        xyz_lines = []
        xyz_lines.append(f"{len(self.symbols)}")
        # Add qcio data to comments line
        comments = f"{' '.join([f'{k}={v}' for k, v in qcio_data.items()])}"
        # Add any other comments
        if xyz_comments := self.extras.get(self._xyz_comment_key, []):
            comments += " " + " ".join(xyz_comments)
        xyz_lines.append(comments)

        # Create a format string using the precision parameter
        format_str = f"{{:2s}} {{: >18.{precision}f}} {{: >18.{precision}f}} {{: >18.{precision}f}}"  # noqa: E501

        for symbol, (x, y, z) in zip(self.symbols, geometry_angstrom):
            xyz_lines.append(format_str.format(symbol, x, y, z))
        xyz_lines.append("")  # Append newline to end of file
        return "\n".join(xyz_lines)

    def __repr_args__(self) -> "ReprArgs":
        """A helper for __repr__ that returns a list of tuples of the form
        (name, value).
        """
        return [  # pragma: no cover
            ("formula", self.formula),
        ]

    def add_smiles(
        self,
        *,
        program: str = "rdkit",
        hydrogens: bool = False,
    ) -> None:
        """Add SMILES data to the identifiers. The SMILES will be generated from the
            structure using the specified program.

        Args:
            program: The program to use to generate the SMILES. Defaults to "rdkit".
            hydrogens: Whether to include hydrogens in the SMILES string. Defaults to
                False.

        Example:
            ```python
            struct.add_smiles()
            struct.ids.smiles
            'CCO'
            struct.ids.canonical_smiles
            'CCO'
            ```
        """
        smiles = self.to_smiles(program=program, hydrogens=hydrogens)
        identifiers = {"smiles": smiles}

        if hydrogens:
            identifiers["canonical_explicit_hydrogen_smiles"] = smiles
        else:
            identifiers["canonical_smiles"] = smiles

        identifiers["canonical_smiles_program"] = program
        self.add_identifiers(identifiers)
        # Ensure pydantic knows the field has been set
        self.__pydantic_fields_set__.add("identifiers")

    def add_identifiers(self, identifiers: Dict[str, str]) -> None:
        """Add an identifier to the structure.

        Args:
            identifiers: The identifiers to add to the structure.

        Example:
            ```python
            struct.add_identifiers({"name": "water"})
            struct.ids.name
            'water'
            ```

            ```python
            struct.add_identifier({"smiles": "CCO"})
            struct.ids.smiles
            'CCO'
            ```
        """
        # Ensure the identifier is valid
        for identifier in identifiers.keys():
            assert hasattr(
                self.identifiers, identifier
            ), f"Invalid identifier: '{identifier}'"

        new_identifiers = self.identifiers.model_copy(update=identifiers)
        object.__setattr__(self, "identifiers", new_identifiers)

    @model_validator(mode="before")
    def _validate_symbols_and_geometry(cls, values):
        """Ensure symbols are valid atomic symbols and geometry is correct."""
        symbols = [symbol.capitalize() for symbol in values.get("symbols", [])]
        for symbol in symbols:
            if not hasattr(pt, symbol):
                raise ValueError(f"Invalid atomic symbol: '{symbol}'")
        values["symbols"] = symbols

        geometry = values.get("geometry")
        if geometry is not None:
            n_atoms = len(values["symbols"])
            values["geometry"] = np.array(geometry).reshape(n_atoms, 3)

        return values

    @field_serializer("connectivity")
    def _serialize_connectivity(self, connectivity, _info) -> List[List[float]]:
        """Serialize connectivity to a list of tuples.

        Cannot have homogeneous data types in .toml files so must cast all values to
        floats.
        """
        return [[float(val) for val in bond] for bond in connectivity]

    @property
    def geometry_angstrom(self) -> np.ndarray:
        """Return the geometry of the structure in Angstrom."""
        return self.geometry * BOHR_TO_ANGSTROM

    @property
    def atomic_numbers(self) -> List[int]:
        """Return the atomic numbers of the atoms in the structure."""
        return [getattr(pt, symbol).number for symbol in self.symbols]

    @property
    def formula(self) -> str:
        """Return the molecular formula of the structure using the Hill System."""
        # https://chemistry.stackexchange.com/questions/1239/order-of-elements-in-a-formula

        counter_elements = Counter(self.symbols)

        carbon_count = counter_elements.pop("C", 0)
        hydrogen_count = counter_elements.pop("H", 0)

        # Sort remaining elements alphabetically
        sorted_elements = sorted(counter_elements.items())

        if hydrogen_count > 0:
            sorted_elements = [("H", hydrogen_count)] + sorted_elements
        if carbon_count > 0:
            sorted_elements = [("C", carbon_count)] + sorted_elements

        return "".join(
            f"{element}{count if count > 1 else ''}"
            for element, count in sorted_elements
        )

    def model_dump(self, **kwargs) -> Dict[str, Any]:
        """Handle tuple in connectivity"""
        as_dict = super().model_dump(**kwargs)
        # Connectivity may be empty and super().dict() will remove empty values
        if (connectivity := as_dict.get("connectivity")) is not None:
            # Must cast all values to floats as toml cannot handle mixed types
            as_dict["connectivity"] = [
                [float(val) for val in bond] for bond in connectivity
            ]
        return as_dict


@renamed_class(Structure)
class Molecule(Structure):
    pass



"""End user output and results objects from a calculation."""

from __future__ import annotations

import sys
import warnings
from itertools import product
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    Union,
    get_args,
)

import numpy as np
from pydantic import field_validator, model_validator
from typing_extensions import Self

from qcio.helper_types import SerializableNDArray

from .base_models import CalcType, Files, Provenance, QCIOModelBase
from .inputs import DualProgramInput, FileInput, Inputs, InputType, ProgramInput
from .structure import Structure
from .utils import deprecated_class

if TYPE_CHECKING:  # pragma: no cover
    pass


__all__ = [
    "SinglePointResults",
    "Wavefunction",
    "OptimizationResults",
    "ProgramOutput",
    "ResultsType",
    "Results",
    "SinglePointOutput",
    "ProgramFailure",
    "OptimizationOutput",
    "ConformerSearchResults",
]


class Wavefunction(QCIOModelBase):
    """The wavefunction for a single point calculation.

    Attributes:
        scf_eigenvalues_a: The SCF alpha-spin orbital eigenvalues.
        scf_eigenvalues_b: The SCF beta-spin orbital eigenvalues.
        scf_occupations_a: The SCF alpha-spin orbital occupations.
        scf_occupations_b: The SCF beta-spin orbital occupations.
    """

    scf_eigenvalues_a: Optional[SerializableNDArray] = None
    scf_eigenvalues_b: Optional[SerializableNDArray] = None
    scf_occupations_a: Optional[SerializableNDArray] = None
    scf_occupations_b: Optional[SerializableNDArray] = None

    @field_validator(
        "scf_eigenvalues_a",
        "scf_eigenvalues_b",
        "scf_occupations_a",
        "scf_occupations_b",
    )
    @classmethod
    def to_numpy(cls, val, _info) -> Optional[np.ndarray]:
        return np.asarray(val) if val is not None else None


class SinglePointResults(Files):
    """The computed results from a single point calculation.

    Attributes:
        calcinfo_natoms: The number of atoms as computed by the program.
        calcinfo_nalpha: The number of alpha electrons as computed by the program.
        calcinfo_nbeta: The number of beta electrons as computed by the program.
        calcinfo_nbasis: The number of basis functions in the calculation.
        calcinfo_nmo: The number of molecular orbitals in the calculation.

        energy: The electronic energy of the structure in `Hartrees`.
        gradient: The gradient of the structure in `Hartrees/Bohr`.
        hessian: The hessian of the structure in `Hartrees/Bohr^2`.
        nuclear_repulsion_energy: The nuclear repulsion energy of the structure in
            Hartrees.

        wavefunction: Wavefunction data from the calculation.

        freqs_wavenumber: The frequencies of the structure in wavenumbers.
        normal_modes_cartesian: 3D n_vibmodes x n_atoms x 3 array containing
            un-mass-weighted Cartesian displacements of each normal mode.
        gibbs_free_energy: Gibbs free energy (i.e. thermochemical analysis) in Hartrees
            of a system where translation / rotation / vibration degrees of freedom are
            approximated using ideal gas / rigid rotor / harmonic oscillator
            respectively.
        scf_dipole_moment: The x, y, z component of the dipole moment of the structure
            in units of e a0 (NOT Debye!).

    """

    # calcinfo contains general information about the calculation
    calcinfo_natoms: Optional[int] = None
    calcinfo_nbasis: Optional[int] = None
    calcinfo_nmo: Optional[int] = None
    calcinfo_nalpha: Optional[int] = None
    calcinfo_nbeta: Optional[int] = None

    # Core properties
    energy: Optional[float] = None
    gradient: Optional[SerializableNDArray] = None  # Coerced to 2D array
    hessian: Optional[SerializableNDArray] = None  # Coerced to 2D array
    nuclear_repulsion_energy: Optional[float] = None

    # Wavefunction data
    wavefunction: Optional[Wavefunction] = None

    # Frequency data
    freqs_wavenumber: List[float] = []
    normal_modes_cartesian: Optional[SerializableNDArray] = None  # Coerced to 3D array
    gibbs_free_energy: Optional[float] = None

    # SCF results
    scf_dipole_moment: Optional[List[float]] = None

    @field_validator("normal_modes_cartesian")
    @classmethod
    def _validate_normal_modes_cartesian_shape(cls, v: SerializableNDArray):
        if v is not None:
            # Assume array has length of the number of normal modes
            n_normal_modes = len(v)
            return np.asarray(v).reshape(n_normal_modes, -1, 3)

    @field_validator("gradient")
    @classmethod
    def _validate_gradient_shape(cls, v: SerializableNDArray):
        """Validate gradient is n x 3"""
        if v is not None:
            return np.asarray(v).reshape(-1, 3)

    @field_validator("hessian")
    @classmethod
    def _validate_hessian_shape(cls, v: SerializableNDArray):
        """Validate hessian is square"""
        if v is not None:
            v = np.asarray(v)
            n = int(np.sqrt(v.size))
            return v.reshape((n, n))

    def return_result(self, calctype: CalcType) -> Union[float, SerializableNDArray]:
        """Return the primary result of the calculation."""
        return getattr(self, calctype.value)

    @model_validator(mode="after")
    def _ensure_results(self) -> Self:
        """Ensure that at least one result is present."""
        if all(
            result is None
            for result in [
                self.energy,
                self.gradient,
                self.hessian,
            ]
        ):
            raise ValueError(
                "SinglePointResults requires either an energy, gradient, or hessian "
                "value."
            )
        return self


class OptimizationResults(Files):
    """Computed properties for an optimization.

    Attributes:
        energies: The energies for each step of the optimization.
        structures: The Structure objects for each step of the optimization.
        final_structure: The final, optimized Structure.
        trajectory: The SinglePointOutput objects for each step of the optimization.
    """

    trajectory: List[
        Union[
            ProgramOutput[ProgramInput, SinglePointResults],
            ProgramOutput[ProgramInput, Files],
        ]
    ] = []

    @property
    def final_structure(self) -> Structure:
        """The final Structure in the optimization."""
        return self.structures[-1]

    @property
    def final_molecule(self) -> Structure:
        warnings.warn(
            ".final_molecule is being depreciated and will be removed in a future. "
            "Please use .final_structure instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.final_structure

    @property
    def final_energy(self) -> Optional[float]:  # Optional for np.nan
        """
        The final energy in the optimization. Is `np.nan` if final calculation failed.
        """
        return self.energies[-1]

    @property
    def energies(self) -> np.ndarray:
        """The energies for each step of the optimization."""
        return np.array(
            [
                # ensure_structured_results_on_success validator ensures .results
                # is not Files if success is True so # type: ignore
                output.results.energy if output.success else np.nan  # type: ignore
                for output in self.trajectory
            ],
            dtype=float,
        )

    @property
    def structures(self) -> List[Structure]:
        """The Structure objects for each step of the optimization."""
        return [output.input_data.structure for output in self.trajectory]

    @property
    def molecules(self) -> List[Structure]:
        """The Structure objects for each step of the optimization."""
        warnings.warn(
            ".molecules is being depreciated and will be removed in a future. "
            "Please use .structures instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.structures

    def return_result(self, calctype: CalcType) -> Optional[Structure]:
        """Return the primary result of the calculation."""
        return self.final_structure

    def __repr_args__(self):
        """Custom repr to avoid printing the entire collection of objects."""
        return [
            ("final_structure", f"{self.final_structure}"),
            ("trajectory", "[...]"),
            ("energies", "[...]"),
            ("structures", "[...]"),
        ]

    def to_xyz(self) -> str:
        """Return the trajectory as an `xyz` string."""
        return "".join(
            prog_output.input_data.structure.to_xyz() for prog_output in self.trajectory
        )

    def save(
        self,
        filepath: Union[Path, str],
        exclude_none: bool = True,
        exclude_unset: bool = True,
        indent: int = 4,
        **kwargs: Dict[str, Any],
    ) -> None:
        """Save an OptimizationOutput to a file.

        Args:
            filepath: The path to save the molecule to.
            exclude_none: If True, attributes with a value of None will not be written
                to the file.
            exclude_unset: If True, attributes that have not been set will not be
                written to the file.
            **kwargs: Additional keyword arguments to pass to the json serializer.

        Note:
            If the filepath has a `.xyz` extension, the trajectory will be saved to a
            multi-structure `xyz` file.
        """
        filepath = Path(filepath)
        if filepath.suffix == ".xyz":
            filepath.write_text(self.to_xyz())
            return
        super().save(filepath, exclude_none, exclude_unset, indent, **kwargs)


class ConformerSearchResults(Files):
    """Results from a conformer search calculation.

    Conformers and rotamers are sorted by energy.

    Attributes:
        conformers: The conformers found in the search.
        conformer_energies: The energies for each conformer.
        rotamers: The rotamers found in the search.
        rotamer_energies: The energies for each rotamer.
    """

    conformers: List[Structure] = []
    conformer_energies: SerializableNDArray = np.array([])
    rotamers: List[Structure] = []
    rotamer_energies: SerializableNDArray = np.array([])

    @model_validator(mode="after")
    def _energies_size(self) -> "ConformerSearchResults":
        """Ensure the energies are the same size as the conformers and rotamers."""
        if self.conformer_energies.size > 0 and (
            self.conformer_energies.size != len(self.conformers)
        ):
            raise ValueError(
                "The number of conformer energies must match the number of conformers."
            )
        if self.rotamer_energies.size > 0 and (
            self.rotamer_energies.size != len(self.rotamers)
        ):
            raise ValueError(
                "The number of rotamer energies must match the number of rotamers."
            )
        return self

    @model_validator(mode="after")
    def _sort_by_energy(self) -> "ConformerSearchResults":
        """Sort conformers and rotamers by energy."""

        # Sort conformers and their energies together
        if self.conformer_energies.size > 0:
            sorted_indices = np.argsort(self.conformer_energies)
            self.conformers[:] = [self.conformers[i] for i in sorted_indices]
            self.conformer_energies[:] = self.conformer_energies[sorted_indices]

        # Sort rotamers and their energies together
        if self.rotamer_energies.size > 0:
            sorted_indices = np.argsort(self.rotamer_energies)
            self.rotamers[:] = [self.rotamers[i] for i in sorted_indices]
            self.rotamer_energies[:] = self.rotamer_energies[sorted_indices]

        return self

    @property
    def conformer_energies_relative(self) -> np.ndarray:
        """The relative energies for each conformer in the search."""
        if self.conformer_energies.size == 0:
            return np.array([])
        return self.conformer_energies - self.conformer_energies.min()

    @property
    def rotamer_energies_relative(self) -> np.ndarray:
        """The relative energies for each rotamer in the search."""
        if self.rotamer_energies.size == 0:
            return np.array([])
        return self.rotamer_energies - self.rotamer_energies.min()


Results = Union[Files, SinglePointResults, OptimizationResults, ConformerSearchResults]
ResultsType = TypeVar("ResultsType", bound=Results)


class ProgramOutput(QCIOModelBase, Generic[InputType, ResultsType]):
    """The core output object from a quantum chemistry calculation.

    Attributes:
        input_data: The input data for the calculation. Any of `qcio.Inputs`.
        success: Whether the calculation was successful.
        results: The results of the calculation. Contains parsed values and files.
            Any of `qcio.Results`.
        stdout: The standard output from the calculation.
        traceback: The traceback from the calculation, if it failed.
        provenance: The provenance information for the calculation.
        extras Dict[str, Any]: Additional information to bundle with the object. Use for
            schema development and scratch space.
        pstdout str: `@property` Print the stdout text.
        ptraceback str: `@property` Print the traceback text.
    """

    input_data: InputType
    success: Literal[True, False]
    results: ResultsType
    stdout: Optional[str] = None
    traceback: Optional[str] = None
    provenance: Provenance

    def __init__(self, **data: Any):
        """Backwards compatibility for files attribute."""

        if "files" in data:
            warnings.warn(
                "The 'files' attribute has been moved to 'results.files'. Please "
                "update your code accordingly.",
                category=FutureWarning,
                stacklevel=2,
            )

            # This moves files from the top level to the results attribute
            if isinstance(data["results"], dict):
                results_files_dict = data["results"].get("files", {})
            else:  # data["results"] is Files, SinglePointResults, OptimizationResults
                results_files_dict = data["results"].files

            results_files_dict.update(**data.pop("files"))

        super().__init__(**data)

    def model_post_init(self, __context) -> None:
        """Parameterize the class (if not set explicitly)."""
        # Check if the current class is still generic, do not override if explicitly set
        if self.__class__ is ProgramOutput:
            input_type = type(self.input_data)
            results_type = type(self.results)
            self.__class__ = ProgramOutput[input_type, results_type]  # type: ignore # noqa 501

    @model_validator(mode="after")
    def ensure_traceback_on_failure(self) -> Self:
        if self.success is False:
            assert (
                self.traceback is not None
            ), "A traceback must be provided for failed calculations."
        return self

    @model_validator(mode="after")
    def _ensure_structured_results_on_success(self) -> Self:
        """Ensure structured results are provided for successful, non FileInputs."""
        # Covers case of ProgramInput and DualProgramInput
        if self.success is True and isinstance(self.input_data, ProgramInput):
            assert type(self.results) is not Files, (
                "Structured results must be provided for successful, non FileInput "
                "calculations."
            )

        return self

    @model_validator(mode="after")
    def ensure_primary_result_on_success(self) -> Self:
        if type(self.results) is SinglePointResults:
            # Ensure the primary calctype result is present
            calctype_val = self.input_data.calctype.value  # type: ignore
            assert (
                getattr(self.results, calctype_val) is not None
            ), f"Missing the primary result: {calctype_val}."

        return self

    @property
    def pstdout(self) -> None:
        """Print the stdout text"""
        print(self.stdout)

    @property
    def ptraceback(self) -> None:
        """Print the traceback text"""
        print(self.traceback)

    def __repr_args__(self) -> List[Tuple[str, Any]]:
        """Exclude stdout and traceback from the repr and ensure success is first"""
        args = super().__repr_args__()

        # Replace stdout and traceback with "<...>"
        filtered_args = [
            (key, value if key not in {"stdout", "traceback"} else "<...>")
            for key, value in args
        ]
        # Ensure success is first
        success_arg = [(key, value) for key, value in filtered_args if key == "success"]
        other_args = [(key, value) for key, value in filtered_args if key != "success"]
        return success_arg + other_args

    @property
    def files(self) -> Dict[str, Union[str, bytes]]:
        """Return the files attribute."""
        # Depreciation warning
        warnings.warn(
            ".files has been moved to .results.files. "
            "Please access it there going forward.",
            category=FutureWarning,
            stacklevel=2,
        )
        return self.results.files

    @property
    def return_result(self) -> Union[float, SerializableNDArray, Optional[Structure]]:
        """Return the primary result of the calculation."""
        warnings.warn(
            ".return_result is being depreciated and will be removed in a future. "
            "Please access results directly at .results instead.",
            category=FutureWarning,
            stacklevel=2,
        )
        # For mypy
        assert self.results is not None, "No results exist on this ProgramOutput"
        assert type(self.input_data) is not FileInput, "FileInputs have no results."
        return self.results.return_result(self.input_data.calctype)  # type: ignore


# For compatibility with old API
@deprecated_class("ProgramOutput[StructuredInputs, OptimizationResults]")
class OptimizationOutput(ProgramOutput[DualProgramInput, OptimizationResults]):
    success: Literal[True] = True
    traceback: Optional[str] = None


@deprecated_class("ProgramOutput[ProgramInput, SinglePointResults]")
class SinglePointOutput(ProgramOutput[ProgramInput, SinglePointResults]):
    success: Literal[True] = True
    traceback: Optional[str] = None


@deprecated_class("ProgramOutput[StructuredInputs, Results]")
class ProgramFailure(ProgramOutput[Inputs, Results]):
    success: Literal[False] = False


# Register the concrete classes for serialization
def _register_program_output_classes():
    """Required so that pickle can find the concrete classes for serialization."""
    for input_type, results_type in product(get_args(Inputs), get_args(Results)):
        _class = ProgramOutput[input_type, results_type]
        name = _class.__name__
        this_module = sys.modules[__name__]
        if name not in this_module.__dict__:
            # Directly declare the type to ensure it is registered
            setattr(this_module, name, _class)


# Call this function during module initialization
_register_program_output_classes()



"""The Base model from which all QCIO Model objects inherit."""

import json
from abc import ABC
from base64 import b64decode, b64encode
from enum import Enum
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

import numpy as np
import toml
import yaml
from pydantic import BaseModel, field_serializer, field_validator
from typing_extensions import Self

from ..helper_types import StrOrPath

if TYPE_CHECKING:  # pragma: no cover
    from pydantic.typing import ReprArgs


__all__ = ["Files", "Provenance", "Model", "CalcType"]


class QCIOModelBase(BaseModel, ABC):
    """Base Model for all QCIO objects.

    Attributes:
        version: The version of the schema.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.
    """

    # NOTE: No version for now until we have a stable schema
    # Or maybe have version be version of qcio that generated the object?
    # version: ClassVar[Literal["v1"]] = "v1"
    extras: Dict[str, Any] = {}

    model_config = {
        # Raises an error if extra fields are passed to model.
        "extra": "forbid",
        # Allow numpy types in models. Pydantic will no longer raise an exception for
        # types it doesn't recognize.
        # https://docs.pydantic.dev/latest/usage/types/#arbitrary-types-allowed
        "arbitrary_types_allowed": True,
        # Don't allow mutation of objects
        # https://docs.pydantic.dev/2.3/api/config/#pydantic.config.ConfigDict.frozen
        "frozen": True,
    }

    @classmethod
    def open(cls, filepath: Union[Path, str]) -> Union[Self, List[Self]]:
        """Instantiate an object from data saved to disk.

        Args:
            filepath: The path to the object on disk.

        Returns:
            The instantiated object or a list of objects if the file contains multiple
            objects (e.g., a multi-structure xyz file).

        Example:
            ```python
            my_obj = MyModel.open("path/to/file.json")
            ```
        """
        filepath = Path(filepath)
        data = filepath.read_text()

        if filepath.suffix in [".yaml", ".yml"]:
            return cls.model_validate(yaml.safe_load(data))
        elif filepath.suffix == ".toml":
            return cls.model_validate(toml.loads(data))

        # Assume json for all other file extensions
        return cls.model_validate_json(data)

    def save(
        self,
        filepath: Union[Path, str],
        exclude_none: bool = True,
        exclude_unset: bool = True,
        indent: int = 4,
        **kwargs,
    ) -> None:
        """
        Save an object to disk as `json`, `yaml`, or `toml`. Objects such as `Structure`
        and `OptimizationResults` can additionally be saved as `xyz` files.

        Note:
            By default the object will be saved as a `json` file. If the file extension
            is `.yaml` or `.yml`, the object will be saved as a `yaml` file. If the file
            extension is `.toml`, the object will be saved as a `toml` file. If the file
            extension is `.xyz`, the object will be saved as an `xyz` file (for objects
            that support this format such as a `Structure` or an `OptimizationResults`
            which contains `.trajectory: list[Structure]`).

            Additionally, padding will be added to the file by default to make it more
            human-readable. You can adjust the amount of padding added by changing the
            `indent` parameter. Pass `indent=None` to create a more compact file.

        Args:
            filepath: The path to write the object to.
            exclude_none: If True, attributes with a value of None will not be written.
                Changing default behavior from pydantic.model_dump() to True.
            exclude_unset: If True, attributes that have not been set will not be
                written (i.e., values set to their default value).
            indent: The number of spaces to use for indentation in the JSON file. 0
                creates a more compact JSON file, 4 is more human-readable.
            **kwargs: Additional keyword arguments to pass to the serialization method.

        Example:
            ```python
            my_obj.save("path/to/file.json")
            ```

            ```python
            my_obj.save("path/to/file.yaml")
            ```

            ```python
            my_obj.save("path/to/file.toml")
            ```

            ```python
            my_obj.save("path/to/file.xyz")
            ```
        """
        filepath = Path(filepath)
        filepath.parent.mkdir(exist_ok=True, parents=True)

        if self.extras:
            # Ensure pydantic knows the field has been set
            self.__pydantic_fields_set__.add("extras")

        model_dict = self.model_dump(
            mode="json",
            exclude_none=exclude_none,
            exclude_unset=exclude_unset,
            **kwargs,
        )

        if filepath.suffix in [".yaml", ".yml"]:
            data = yaml.dump(model_dict, indent=indent)

        elif filepath.suffix == ".toml":
            data = toml.dumps(model_dict)

        else:
            # Write data to json regardless of file extension
            data = json.dumps(model_dict, indent=indent)

        filepath.write_text(data)

    def __repr_args__(self) -> "ReprArgs":
        """Only show non empty fields in repr but always show success even if false."""

        def exists(value):
            if isinstance(value, np.ndarray):
                return value.size > 0
            return bool(value)

        return [  # pragma: no cover
            (name, value)
            for name, value in self.__dict__.items()
            if name == "success" or exists(value)
        ]

    def __eq__(self, other: Any) -> bool:
        """Check equality of two objects.

        Necessary because BaseModel.__eq__ does not compare numpy arrays.
        """
        if isinstance(other, self.__class__):
            return self.model_dump() == other.model_dump()
        return False


class Files(QCIOModelBase):
    """File model for handling string and binary data.

    Binary data is encoded as base64 strings during serialization.

    Attributes:
        files: A dict mapping filename to str or bytes data.
    """

    files: Dict[str, Union[str, bytes]] = {}

    @field_validator("files")
    def _convert_base64_to_bytes(cls, value):
        """Convert base64 encoded data to bytes."""
        for filename, data in value.items():
            if isinstance(data, str) and data.startswith("base64:"):
                value[filename] = b64decode(data[7:])
        return value

    @field_serializer("files")
    def _serialize_files(self, files, _info) -> Dict[str, str]:
        """Serialize files to a dict of filename to base64 encoded string."""
        return {
            filename: (
                f"base64:{b64encode(data).decode('utf-8')}"
                if isinstance(data, bytes)
                else data
            )
            for filename, data in files.items()
        }

    def add_file(
        self, filepath: Union[Path, str], relative_dir: Optional[Path] = None
    ) -> None:
        """Add a file to the object. The file will be added at to the `files` attribute
            with the filename as the key and the file data as the value.

        Args:
            filepath: The path to the file.
            relative_dir: The directory to make the file relative to. Helpful when
                adding files from a subdirectory.

        Example:
            ```python
                my_obj.add_file("path/to/file.txt")
                print(my_obj.files)
                # Output: {"file.txt": "file data"}
            ```
        """
        filepath = Path(filepath)
        raw_bytes = filepath.read_bytes()
        try:
            data: Union[str, bytes] = raw_bytes.decode("utf-8")  # str
        except UnicodeDecodeError:
            data = raw_bytes  # bytes

        # Set filename relative to relative_dir
        if relative_dir:
            filename = str(filepath.relative_to(relative_dir))
        else:
            filename = filepath.name

        self.files[filename] = data
        # Add files to __pydantic_fields_set__ to ensure they are included in .save()
        self.__pydantic_fields_set__.add("files")

    def add_files(
        self,
        directory: StrOrPath,
        recursive: bool = False,
        exclude: Optional[List[str]] = None,
    ) -> None:
        """Add all files in a directory to the object.

        Args:
            directory: The directory to add files from.
            recursive: Whether to recursively add files from subdirectories.
            exclude: A list of filenames to exclude from the directory.
        """
        directory = Path(directory)
        if exclude is None:
            exclude = []
        if recursive:
            files = directory.rglob("*")
        else:
            files = directory.glob("*")
        for filepath in files:
            if filepath.is_file() and filepath.name not in exclude:
                self.add_file(filepath, directory)

    def save_files(self, directory: StrOrPath = Path(".")) -> None:
        """Write all files to the specified directory"""
        directory = Path(directory)
        directory.mkdir(exist_ok=True)
        for filename, data in self.files.items():
            mode = "w" if isinstance(data, str) else "wb"
            filepath = directory / filename
            # In case filename is a relative path, create the parent directories
            filepath.parent.mkdir(exist_ok=True, parents=True)
            filepath.open(mode).write(data)

    def __repr_args__(self) -> "ReprArgs":
        """Replace file data with '<bytes>' or '<str>' in __repr__."""
        rargs = super().__repr_args__()
        return [
            (
                (k, v)
                if k != "files"
                else (
                    k,
                    {
                        name: bytes if isinstance(data, bytes) else str
                        for name, data in v.items()
                    },
                )
            )
            for k, v in rargs
        ]


class Provenance(QCIOModelBase):
    """Provenance information for a QC program.

    Attributes:
        program: The name of the program that created the output.
        version: The version of the program that created the output.
        scratch_dir: The working directory used by the program.
        wall_time: The wall time used by the program.
        hostname: The hostname of the machine the program was run on.
        hostcpus: The number of logical CPUs on the host machine where the program ran.
        hostmem: The amount of memory on the host machine where the program ran in GiB.
        extras: Additional information to bundle with the object. Use for schema
            development and scratch space.

    """

    program: str
    program_version: Optional[str] = None
    scratch_dir: Optional[StrOrPath] = None
    wall_time: Optional[float] = None
    hostname: Optional[str] = None
    hostcpus: Optional[int] = None
    hostmem: Optional[int] = None


class CalcType(str, Enum):
    """The Calculation type.

    Attributes:
        energy: Energy calculation.
        gradient: Gradient calculation.
        hessian: Hessian calculation.
        optimization: Geometry optimization (to a minima).
        transition_state: Transition state optimization (to a saddle point).
        conformer_search: Conformer search.
    """

    energy = "energy"
    gradient = "gradient"
    hessian = "hessian"
    optimization = "optimization"
    transition_state = "transition_state"
    conformer_search = "conformer_search"

    def __repr__(self) -> str:
        """Custom repr for CalcType"""
        return f"'{self.name}'"  # pragma: no cover


class Model(QCIOModelBase):
    """The model for the quantum chemistry calculation.

    Attributes:
        method: The name of the method to be used in the calculation. Named according to
            the convention of the program being called. If an MM calculation then the
            name of the force field.
        basis: The name of the basis set to be used in the calculation. Named according
            to the convention of the program being called.
    """

    method: str
    basis: Optional[str] = None
